/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubelib.block

import codechicken.lib.render.particle.CustomParticleHandler
import de.keri.cubelib.client.render.RenderingRegistry
import de.keri.cubelib.client.render.block.{IBlockColorHandler, RenderBlocks}
import de.keri.cubelib.client.texture.ITextureBlock
import de.keri.cubelib.item.ItemBlockBase
import de.keri.cubelib.proxy.ClientProxy
import de.keri.cubelib.util.MetadataUtils
import de.keri.cubeloader.loader.ILoadable
import de.keri.cubeloader.loader.loadstage.{EnumStageType, LoadStage}
import net.minecraft.block.material.{MapColor, Material}
import net.minecraft.block.state.{BlockStateContainer, IBlockState}
import net.minecraft.block.{Block, ITileEntityProvider}
import net.minecraft.client.particle.ParticleManager
import net.minecraft.client.renderer.texture.{TextureAtlasSprite, TextureMap}
import net.minecraft.creativetab.CreativeTabs
import net.minecraft.entity.EntityLivingBase
import net.minecraft.item.{ItemBlock, ItemStack}
import net.minecraft.tileentity.TileEntity
import net.minecraft.util._
import net.minecraft.util.math.{BlockPos, RayTraceResult}
import net.minecraft.world.{IBlockAccess, World, WorldServer}
import net.minecraftforge.fml.common.registry.ForgeRegistries
import net.minecraftforge.fml.relauncher.{Side, SideOnly}

/**
  * A block class with support for metadata based subtypes and default implementations
  * of ILoadable, ITextureBlock, IBlockColorHandler and IBlockRenderTypeProvider.
  *
  * @param blockName The name of the block
  * @param material  Material of the block
  * @param mapColor  MapColor of the block
  * @param subNames  All metadata based subtypes to be added for this block
  */
class BlockBase(blockName: String, material: Material, mapColor: MapColor, subNames: Array[String]) extends Block(material, mapColor)
  with ILoadable with ITileEntityProvider with ITextureBlock with IBlockColorHandler {

  setDefaultState(getBlockState.getBaseState.withProperty[Integer, Integer](CommonProperties.META_STATE, 0))

  private var modid: String = _
  @SideOnly(Side.CLIENT)
  private var textures: Array[TextureAtlasSprite] = _

  /**
    * @param blockName The name of the block
    * @param material  Material of the block
    * @param mapColor  MapColor of the block
    */
  def this(blockName: String, material: Material, mapColor: MapColor) = this(blockName, material, mapColor, null)

  /**
    * @param blockName The name of the block
    * @param material  Material of the block
    * @param subNames  All metadata based subtypes to be added for this block
    */
  def this(blockName: String, material: Material, subNames: Array[String]) = this(blockName, material, MapColor.BLACK, subNames)

  /**
    * @param blockName The name of the block
    * @param material  Material of the block
    */
  def this(blockName: String, material: Material) = this(blockName, material, MapColor.BLACK, null)

  override def createNewTileEntity(world: World, meta: Int): TileEntity = null

  override def createBlockState: BlockStateContainer = new BlockStateContainer(this, CommonProperties.META_STATE)

  override def getMetaFromState(state: IBlockState): Int = state.getValue(CommonProperties.META_STATE).intValue()

  override def getStateFromMeta(meta: Int): IBlockState = getDefaultState.withProperty[Integer, Integer](CommonProperties.META_STATE, meta)

  override def getSubBlocks(tab: CreativeTabs, items: NonNullList[ItemStack]): Unit = if (tab == getCreativeTab) {
    if (getSubNames != null) for (i <- subNames.indices) items.add(new ItemStack(this, 1, i))
    else items.add(new ItemStack(this, 1, 0))
  }

  @SideOnly(Side.CLIENT)
  override def getRenderType(state: IBlockState): EnumBlockRenderType = RenderBlocks.DEFAULT

  @SideOnly(Side.CLIENT)
  override def addDestroyEffects(world: World, pos: BlockPos, manager: ParticleManager): Boolean = {
    CustomParticleHandler.handleDestroyEffects(world, pos, manager)
  }

  @SideOnly(Side.CLIENT)
  override def addHitEffects(state: IBlockState, world: World, target: RayTraceResult, manager: ParticleManager): Boolean = {
    CustomParticleHandler.handleHitEffects(state, world, target, manager)
  }

  override def addLandingEffects(state: IBlockState, world: WorldServer, pos: BlockPos, state2: IBlockState, entity: EntityLivingBase, amount: Int): Boolean = {
    CustomParticleHandler.handleLandingEffects(world, pos, entity, amount)
  }

  @SideOnly(Side.CLIENT)
  override def registerTextures(map: TextureMap): Unit = {
    if (subNames != null) {
      textures = Array.ofDim[TextureAtlasSprite](subNames.length)
      for (i <- subNames.indices) textures(i) = map.registerSprite(new ResourceLocation(modid, s"blocks/${blockName}_${subNames(i)}"))
    }
    else {
      textures = Array.ofDim[TextureAtlasSprite](1)
      textures(0) = map.registerSprite(new ResourceLocation(modid, s"blocks/$blockName"))
    }
  }

  @SideOnly(Side.CLIENT)
  override def getTexture(meta: Int, side: Int): TextureAtlasSprite = textures(meta)

  @SideOnly(Side.CLIENT)
  override def getTexture(world: IBlockAccess, pos: BlockPos, side: Int): TextureAtlasSprite = getTexture(MetadataUtils.getBlockMetadata(world, pos), side)

  @SideOnly(Side.CLIENT)
  override def getColorMultiplier(meta: Int, side: Int): Int = 0xFFFFFFFF

  @SideOnly(Side.CLIENT)
  override def getColorMultiplier(world: IBlockAccess, pos: BlockPos, side: Int) = getColorMultiplier(MetadataUtils.getBlockMetadata(world, pos), side)

  override def load(loadStage: LoadStage): Unit = {
    loadStage.getStageType match {
      case EnumStageType.PRE_INIT => {
        modid = loadStage.getModMetadata.getModid
        setRegistryName(modid, blockName)
        setUnlocalizedName(s"$modid.$blockName")
        ForgeRegistries.BLOCKS.register(this)
        ForgeRegistries.ITEMS.register(getItemBlock.setRegistryName(getRegistryName))
      }
      case _ =>
    }
  }

  @SideOnly(Side.CLIENT)
  override def loadClient(loadStage: LoadStage): Unit = {
    loadStage.getStageType match {
      case EnumStageType.PRE_INIT => {
        ClientProxy.getTextureRegister.registerHandler(this)
        RenderingRegistry.registerBlock(this)
      }
      case _ =>
    }
  }

  @SideOnly(Side.SERVER)
  override def loadServer(loadStage: LoadStage): Unit = {}

  /**
    * Getter method for retrieving the modid this block is registered to
    *
    * @return The modid this block is registered to
    */
  def getModid: String = modid

  /**
    * Getter method for retrieving the name this block is registered to
    *
    * @return The name this block is registered to
    */
  def getBlockName: String = blockName

  /**
    * Getter method for retrieving all metadata based subtypes of this block
    * as an array of Strings
    *
    * @return The metadata based subtypes of this block
    */
  def getSubNames: Array[String] = subNames

  /**
    * Getter method for retrieving the ItemBlock that is registered for this block
    *
    * @return The ItemBlock that is registered for this block
    */
  def getItemBlock: ItemBlock = new ItemBlockBase(this)

  /**
    * Getter method for retrieving the CreativeTabs this block is put in
    *
    * @return The CreativeTabs this block is put in
    */
  def getCreativeTab: CreativeTabs = CreativeTabs.BUILDING_BLOCKS

}
