/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubelib.client.render

import codechicken.lib.model.ModelRegistryHelper
import codechicken.lib.render.block.BlockRenderingRegistry
import com.google.common.collect.Lists
import de.keri.cubelib.client.render.block.{BlockRenderingAdapter, IBlockRenderingHandler}
import de.keri.cubelib.client.render.item.{EnumItemRenderType, IItemRenderTypeProvider, IItemRenderingHandler, ItemRenderingAdapter}
import de.keri.cubelib.exception.RegistryException
import net.minecraft.block.Block
import net.minecraft.client.renderer.ItemMeshDefinition
import net.minecraft.client.renderer.block.model.ModelResourceLocation
import net.minecraft.client.renderer.block.statemap.StateMap
import net.minecraft.item.{Item, ItemStack}
import net.minecraft.util.EnumBlockRenderType
import net.minecraftforge.client.model.ModelLoader
import net.minecraftforge.common.util.EnumHelper

import scala.collection.mutable.HashMap

object RenderingRegistry {

  private var BLOCK_RENDER_ID = 0
  private var ITEM_RENDER_ID = 0
  private val BLOCK_RENDERING_HANDLERS = new HashMap[EnumBlockRenderType, IBlockRenderingHandler]()
  private val ITEM_RENDERING_HANDLERS = new HashMap[EnumItemRenderType, IItemRenderingHandler]()

  def isTypeRegistered(renderType: EnumBlockRenderType): Boolean = BLOCK_RENDERING_HANDLERS.contains(renderType)

  def isTypeRegistered(renderType: EnumItemRenderType): Boolean = ITEM_RENDERING_HANDLERS.contains(renderType)

  /**
    * Gets the next available EnumBlockRenderType based on an int
    * that's increased every time this method is called.
    *
    * @return the next available EnumBlockRenderType
    */
  def getNextAvailableBlockRenderType: EnumBlockRenderType = {
    val renderType = BlockRenderingRegistry.createRenderType(s"ibrh_$BLOCK_RENDER_ID")
    BLOCK_RENDER_ID += 1
    return renderType
  }

  /**
    * Gets the next available EnumItemRenderType based on an int
    * that's increased every time this method is called.
    *
    * @return the next available EnumItemRenderType
    */
  def getNextAvailableItemRenderType: EnumItemRenderType = {
    val renderType = EnumHelper.addEnum(classOf[EnumItemRenderType], s"iirh_$ITEM_RENDER_ID", Array.ofDim[Class[_]](0))
    ITEM_RENDER_ID += 1
    return renderType
  }

  /**
    * Registers an IBlockRenderingHandler.
    *
    * @param handler the IBlockRenderingHandler instance to be registered
    */
  def registerBlockRenderingHandler(handler: IBlockRenderingHandler): Unit = {
    val renderType = handler.getRenderType
    BlockRenderingRegistry.registerRenderer(renderType, new BlockRenderingAdapter(handler))
    BLOCK_RENDERING_HANDLERS.put(handler.getRenderType, handler)
  }

  /**
    * Registers an IItemRenderingHandler.
    *
    * @param handler the IItemRenderingHandler instance to be registered
    */
  def registerItemRenderingHandler(handler: IItemRenderingHandler): Unit = {
    ITEM_RENDERING_HANDLERS.put(handler.getRenderType, handler)
  }

  /**
    * Registers a Block and builds a StateMap to ignore all IProperties in
    * the BlockStateContainer. Also registers a dummy IBakedModel for handling
    * particles and to make the ModelLoader shut up.
    *
    * @param block the Block to be registered
    */
  def registerBlock(block: Block): Unit = {
    val renderType = block.getRenderType(block.getDefaultState)
    val adapter = new BlockRenderingAdapter(BLOCK_RENDERING_HANDLERS.apply(renderType))
    val item = Item.getItemFromBlock(block)
    val location = new ModelResourceLocation(block.getRegistryName.toString)
    val locationInventory = new ModelResourceLocation(block.getRegistryName, "inventory")
    ModelLoader.setCustomStateMapper(block, buildStateMap(block))
    ModelRegistryHelper.register(location, new SimpleBakedModel(Lists.newArrayList()))

    ModelLoader.setCustomMeshDefinition(item, new ItemMeshDefinition {
      override def getModelLocation(stack: ItemStack): ModelResourceLocation = locationInventory
    })

    ModelRegistryHelper.register(locationInventory, adapter)
  }

  /**
    * Registers an Item and registers a dummy IBakedModel to make the
    * ModelLoader shut up.
    *
    * @param item the Itm to be registered
    */
  def registerItem(item: Item): Unit = {
    if (!item.isInstanceOf[IItemRenderTypeProvider]) throw new RegistryException("Item must be an instance of IItemRenderTypeProvider!")
    val renderType = item.asInstanceOf[IItemRenderTypeProvider].getRenderType
    val adapter = new ItemRenderingAdapter(ITEM_RENDERING_HANDLERS.apply(renderType))
    val location = new ModelResourceLocation(item.getRegistryName, "inventory")

    ModelLoader.setCustomMeshDefinition(item, new ItemMeshDefinition {
      override def getModelLocation(stack: ItemStack): ModelResourceLocation = location
    })

    ModelRegistryHelper.register(location, adapter)
  }

  private def buildStateMap(block: Block): StateMap = {
    val builder = new StateMap.Builder
    val container = block.getBlockState

    if (container != null) {
      val properties = container.getProperties

      if (properties != null && !properties.isEmpty) {
        val iterator = properties.iterator()
        while (iterator.hasNext) builder.ignore(iterator.next())
      }
    }

    return builder.build()
  }

}
