/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubelib.client.render.block

import codechicken.lib.render.buffer.BakingVertexBuffer
import codechicken.lib.render.{CCModel, CCRenderState}
import codechicken.lib.vec.uv.IconTransformation
import codechicken.lib.vec.{Cuboid6, Translation, Vector3}
import de.keri.cubelib.client.texture.ITextureBlock
import de.keri.cubelib.util.RenderUtils
import net.minecraft.block.Block
import net.minecraft.block.state.IBlockState
import net.minecraft.client.renderer.BufferBuilder
import net.minecraft.client.renderer.block.model.ItemCameraTransforms
import net.minecraft.client.renderer.texture.TextureAtlasSprite
import net.minecraft.client.renderer.vertex.DefaultVertexFormats
import net.minecraft.item.ItemStack
import net.minecraft.util.math.BlockPos
import net.minecraft.util.{EnumBlockRenderType, EnumFacing}
import net.minecraft.world.IBlockAccess
import net.minecraftforge.fml.relauncher.{Side, SideOnly}
import org.lwjgl.opengl.GL11

@SideOnly(Side.CLIENT)
class RenderBlockDefault extends IBlockRenderingHandler {

  override def renderWorld(world: IBlockAccess, pos: BlockPos, state: IBlockState, buffer: BufferBuilder): Boolean = {
    val model = RenderBlockDefault.MODEL.copy()
    val textureHandler = state.getBlock.asInstanceOf[ITextureBlock]
    val colorHandler = state.getBlock.asInstanceOf[IBlockColorHandler]
    val parentBuffer = BakingVertexBuffer.create()
    parentBuffer.begin(GL11.GL_QUADS, DefaultVertexFormats.ITEM)
    val renderState = CCRenderState.instance()
    renderState.reset()
    renderState.bind(parentBuffer)

    state.getBlock match {
      case block: ICTMBlock => {
        val context = new ConnectedRenderContext
        context.setBlockAccess(world)
        context.setCurrentBlockState(world.getBlockState(pos))

        for (side <- EnumFacing.VALUES.indices) {
          if (block.shouldRenderConnected(world, pos, side)) {
            val texture = block.getConnectedTexture(world, pos, side)
            context.renderStandardBlock(pos, texture.getTextures)
          }
        }

        context.getModel.render(renderState)
      }
      case _ => for (side <- EnumFacing.VALUES.indices) {
        val texture = textureHandler.getTexture(world, pos, side)
        val color = colorHandler.getColorMultiplier(world, pos, side)
        model.setColour(color)
        model.render(renderState, 4 * side, 4 + side * 4, new IconTransformation(texture))
      }
    }

    parentBuffer.finishDrawing()
    return RenderUtils.renderQuads(buffer, world, pos, state, parentBuffer.bake())
  }

  override def renderDamage(world: IBlockAccess, pos: BlockPos, state: IBlockState, texture: TextureAtlasSprite, buffer: BufferBuilder): Unit = {
    val model = RenderBlockDefault.MODEL.copy().apply(new Translation(Vector3.fromBlockPos(pos)))
    val renderState = CCRenderState.instance()
    renderState.reset()
    renderState.bind(buffer)
    model.render(renderState, new IconTransformation(texture))
  }

  override def renderInventory(stack: ItemStack, transformType: ItemCameraTransforms.TransformType, buffer: BufferBuilder): Unit = {
    val model = RenderBlockDefault.MODEL.copy()
    val block = Block.getBlockFromItem(stack.getItem)
    val textureHandler = block.asInstanceOf[ITextureBlock]
    val colorHandler = block.asInstanceOf[IBlockColorHandler]
    val meta = stack.getMetadata
    val renderState = CCRenderState.instance()
    renderState.reset()
    renderState.bind(buffer)

    for (side <- EnumFacing.VALUES.indices) {
      val texture = textureHandler.getTexture(meta, side)
      val color = colorHandler.getColorMultiplier(stack.getMetadata, side)
      model.setColour(color)
      model.render(renderState, new IconTransformation(texture))
    }
  }

  override def getRenderType: EnumBlockRenderType = RenderBlocks.DEFAULT

}

object RenderBlockDefault {

  private final val MODEL = CCModel.quadModel(24).generateBlock(0, new Cuboid6(0D, 0D, 0D, 1D, 1D, 1D)).computeNormals()

}
