/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubelib.item

import net.minecraft.client.renderer.texture.{TextureAtlasSprite, TextureMap}
import net.minecraft.creativetab.CreativeTabs
import net.minecraft.item.{Item, ItemStack}
import net.minecraft.util.{NonNullList, ResourceLocation}
import net.minecraftforge.fml.relauncher.{Side, SideOnly}

/**
  * An Item class with support for metadata based subtypes and default implementations
  * of ILoadable, ITextureItem and IItemRenderTypeProvider.
  *
  * @param itemName The name of the Item
  * @param subNames All metadata based subtypes to be added for this Item
  */
class ItemBase(itemName: String, subNames: Array[String]) extends Item with TDefaultItemImpl {

  setHasSubtypes(subNames != null)

  private var modid: String = _
  @SideOnly(Side.CLIENT)
  private var textures: Array[TextureAtlasSprite] = _

  /**
    * An Item class with support for metadata based subtypes and default implementations
    * of ILoadable, ITextureItem and IItemRenderTypeProvider.
    *
    * @param itemName The name of the Item
    */
  def this(itemName: String) = this(itemName, null)

  override def getMetadata(damage: Int): Int = if (subNames != null) damage else super.getMetadata(damage)

  override def getSubItems(tab: CreativeTabs, items: NonNullList[ItemStack]): Unit = if (tab == getCreativeTab) {
    if (subNames != null) for (i <- subNames.indices) items.add(new ItemStack(this, 1, i))
    else items.add(new ItemStack(this, 1, 0))
  }

  override def getUnlocalizedName(stack: ItemStack): String = {
    if (subNames != null) s"${getUnlocalizedName()}.${subNames(stack.getMetadata)}"
    else getUnlocalizedName()
  }

  override def getCreativeTab: CreativeTabs = CreativeTabs.MISC

  @SideOnly(Side.CLIENT)
  override def registerTextures(map: TextureMap): Unit = {
    if (subNames != null) {
      textures = Array.ofDim[TextureAtlasSprite](subNames.length)
      for (i <- subNames.indices) textures(i) = map.registerSprite(new ResourceLocation(modid, s"items/${itemName}_${subNames(i)}"))
    }
    else {
      textures = Array.ofDim[TextureAtlasSprite](1)
      textures(0) = map.registerSprite(new ResourceLocation(modid, s"items/$itemName"))
    }
  }

  @SideOnly(Side.CLIENT)
  override def getTexture(meta: Int): TextureAtlasSprite = textures(meta)

  @SideOnly(Side.CLIENT)
  override def getTexture(stack: ItemStack): TextureAtlasSprite = null

  override def getModid: String = modid

  override def getItemName: String = itemName

  override def setModid(modid: String): Unit = this.modid = modid

  /**
    * Getter method for retrieving the metadata based subtypes of this Item
    * as an array of Strings.
    * @return the metadata based subtypes of this Item
    */
  def getSubNames: Array[String] = subNames

}
