/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubelib.util

import java.io.InputStreamReader

import codechicken.lib.colour.{Colour, ColourRGBA}
import codechicken.lib.math.MathHelper
import codechicken.lib.vec.uv.UVTranslation
import codechicken.lib.vec.{Rotation, Vector3}
import com.google.gson.{Gson, JsonArray, JsonObject, JsonPrimitive}
import net.minecraft.block.Block
import net.minecraft.item.{Item, ItemStack}
import net.minecraft.util.ResourceLocation

object JsonUtils {

  private final val gson = new Gson();

  def loadJson(location: ResourceLocation): JsonObject = {
    val is = getClass.getResourceAsStream(s"/assets/${location.getResourceDomain}/${location.getResourcePath}.json")
    val isr = new InputStreamReader(is)
    gson.fromJson[JsonObject](isr, classOf[JsonObject])
  }

  def getJsonObject(jsonObject: JsonObject, key: String): JsonObject = jsonObject.get(key).getAsJsonObject

  def getJsonPrimitive(jsonObject: JsonObject, key: String): JsonPrimitive = jsonObject.get(key).getAsJsonPrimitive

  def getJsonArray(jsonObject: JsonObject, key: String): JsonArray = jsonObject.get(key).getAsJsonArray

  def getString(jsonObject: JsonObject, key: String): String = getJsonPrimitive(jsonObject, key).getAsString

  def getInt(jsonObject: JsonObject, key: String): Int = getJsonPrimitive(jsonObject, key).getAsInt

  def getFloat(jsonObject: JsonObject, key: String): Float = getJsonPrimitive(jsonObject, key).getAsFloat

  def getDouble(jsonObject: JsonObject, key: String): Double = getJsonPrimitive(jsonObject, key).getAsDouble

  def getVector3(jsonObject: JsonObject, key: String): Vector3 = {
    val array = getJsonArray(jsonObject, key)
    val x = array.get(0).getAsJsonPrimitive.getAsDouble
    val y = array.get(1).getAsJsonPrimitive.getAsDouble
    val z = array.get(2).getAsJsonPrimitive.getAsDouble
    new Vector3(x, y, z)
  }

  def getUvTranslation(jsonObject: JsonObject, key: String): UVTranslation = {
    val array = getJsonArray(jsonObject, key)
    val u = array.get(0).getAsJsonPrimitive.getAsDouble
    val v = array.get(1).getAsJsonPrimitive.getAsDouble
    return new UVTranslation(u, v)
  }

  def getColor(jsonObject: JsonObject, key: String): Colour = {
    jsonObject.get(key) match {
      case jsonArray: JsonArray => jsonArray.size() match {
        case 3 => {
          val r = jsonArray.get(0).getAsJsonPrimitive.getAsInt
          val g = jsonArray.get(1).getAsJsonPrimitive.getAsInt
          val b = jsonArray.get(2).getAsJsonPrimitive.getAsInt
          new ColourRGBA(r, g, b, 255)
        }
        case 4 => {
          val r = jsonArray.get(0).getAsJsonPrimitive.getAsInt
          val g = jsonArray.get(1).getAsJsonPrimitive.getAsInt
          val b = jsonArray.get(2).getAsJsonPrimitive.getAsInt
          val a = jsonArray.get(3).getAsJsonPrimitive.getAsInt
          new ColourRGBA(r, g, b, a)
        }
        case _ =>
      }
      case _ =>
    }

    null
  }

  def getRotation(jsonObject: JsonObject, key: String): Rotation = {
    val array = jsonObject.get(key).getAsJsonArray

    array.size() match {
      case 4 => {
        val aX = array.get(0).getAsJsonPrimitive.getAsDouble
        val aY = array.get(1).getAsJsonPrimitive.getAsDouble
        val aZ = array.get(2).getAsJsonPrimitive.getAsDouble
        val angle = array.get(3).getAsJsonPrimitive.getAsDouble
        new Rotation(angle * MathHelper.todeg, new Vector3(aX, aY, aZ)).at(Vector3.center).asInstanceOf[Rotation]
      }
      case 7 => {
        val aX = array.get(0).getAsJsonPrimitive.getAsDouble
        val aY = array.get(1).getAsJsonPrimitive.getAsDouble
        val aZ = array.get(2).getAsJsonPrimitive.getAsDouble
        val angle = array.get(3).getAsJsonPrimitive.getAsDouble
        val oX = array.get(4).getAsJsonPrimitive.getAsDouble
        val oY = array.get(5).getAsJsonPrimitive.getAsDouble
        val oZ = array.get(6).getAsJsonPrimitive.getAsDouble
        new Rotation(angle * MathHelper.todeg, new Vector3(aX, aY, aZ)).at(new Vector3(oX, oY, oZ)).asInstanceOf[Rotation]
      }
      case _ =>
    }

    null
  }

  def getItemStack(jsonObject: JsonObject, key: String): ItemStack = {
    val name = jsonObject.get(key).getAsJsonPrimitive.getAsString
    if(Block.getBlockFromName(name) != null) new ItemStack(Block.getBlockFromName(name), 1, 0)
    else new ItemStack(Item.getByNameOrId(name))
  }

}
