/*
 * This file is part of CubeLoader.
 * Copyright (c) 2016 - 2017, KitsuneAlex, All rights reserved.
 *
 * CubeLoader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CubeLoader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with CubeLoader.  If not, see <http://www.gnu.org/licenses/lgpl>.
 */

package de.keri.cubeloader.loader

import java.net.URI
import java.util.UUID

import com.mojang.authlib.GameProfile
import de.keri.cubelib.util.JsonUtils._
import de.keri.cubelib.util.ScalaAdapters._
import de.keri.cubeloader.CubeLoader
import de.keri.cubeloader.loader.loadstage.{EnumStageType, LoadStage}
import de.keri.cubeloader.util.CubeLogger
import net.minecraft.util.ResourceLocation
import net.minecraftforge.fml.common.discovery.ASMDataTable
import net.minecraftforge.fml.common.event._
import net.minecraftforge.fml.common.{Loader, ModContainer, ProgressManager}
import net.minecraftforge.fml.relauncher.{Side, SideOnly}

import scala.collection.mutable.{ArrayBuffer, HashMap}

object ModLoader {

  private final val MODS = new HashMap[String, IMod]()
  private final val LOADED_MODS = new HashMap[String, IMod]()
  private final val METADATA = new HashMap[String, ModMetadata]()
  private final val CONTAINERS = new HashMap[String, ModContainer]()
  private final val LOADABLES = new HashMap[String, ArrayBuffer[ILoadable]]()

  def getMetadata(modid: String): ModMetadata = METADATA.apply(modid)

  def isModLoaded(modid: String): Boolean = LOADED_MODS.contains(modid)

  def getLoadedMods: ArrayBuffer[IMod] = {
    val buffer = new ArrayBuffer[IMod]()
    LOADED_MODS.foreach(m => buffer += m._2)
    return buffer
  }

  def handleConstructionCommon(event: FMLConstructionEvent): Unit = {
    CubeLogger.info("Discovering mods...")
    discoverMods(event.getASMHarvestedData).foreach(m => MODS.put(m.getModid, m))
    CubeLogger.info(s"Done discovering mods! Found ${MODS.size} mod(s) to load")
    MODS.foreach(m => METADATA += loadMetadataJson(m._1))
    METADATA.foreach(m => {
      val container = new CubeModContainer(m._2)
      Loader.instance().setActiveModContainer(container)
      CubeLoader.PROXY.registerContainer(container)
      CONTAINERS.put(m._1, container)
    })
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handlePreInitCommon(event: FMLPreInitializationEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handleInitCommon(event: FMLInitializationEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handlePostInitCommon(event: FMLPostInitializationEvent): Unit = {
    loadMods(event, 0, new ILoadCallback {
      override def onLoad(loadStage: LoadStage): Unit = {
        val meta = loadStage.getModMetadata
        val mod = MODS.apply(meta.getModid)
        LOADED_MODS.put(meta.getModid, mod)
      }
    })

    dispatchLoadables(event, 0, null)
  }

  def handleServerAboutToStartCommon(event: FMLServerAboutToStartEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handleServerStartingCommon(event: FMLServerStartingEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handleServerStartedCommon(event: FMLServerStartedEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handleServerStoppingCommon(event: FMLServerStoppingEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  def handleServerStoppedCommon(event: FMLServerStoppedEvent): Unit = {
    loadMods(event, 0, null)
    dispatchLoadables(event, 0, null)
  }

  @SideOnly(Side.CLIENT)
  def handleConstructionClient(event: FMLConstructionEvent): Unit = {
    var progressBar = ProgressManager.push("[CubeLoader] Loading Mods: Construction", MODS.size)

    loadMods(event, 1, new ILoadCallback {
      override def onLoad(loadStage: LoadStage): Unit = progressBar.step(loadStage.getModMetadata.getName)
    })

    ProgressManager.pop(progressBar)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handlePreInitClient(event: FMLPreInitializationEvent): Unit = {
    var progressBar = ProgressManager.push("[CubeLoader] Loading Mods: PreInit", MODS.size)

    loadMods(event, 1, new ILoadCallback {
      override def onLoad(loadStage: LoadStage): Unit = {
        progressBar.step(loadStage.getModMetadata.getName)
      }
    })

    ProgressManager.pop(progressBar)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleInitClient(event: FMLInitializationEvent): Unit = {
    val progressBar = ProgressManager.push("[CubeLoader] Loading Mods: Init", MODS.size)

    loadMods(event, 1, new ILoadCallback {
      override def onLoad(loadStage: LoadStage): Unit = progressBar.step(loadStage.getModMetadata.getName)
    })

    ProgressManager.pop(progressBar)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handlePostInitClient(event: FMLPostInitializationEvent): Unit = {
    val progressBar = ProgressManager.push("[CubeLoader] Loading Mods: PostInit", MODS.size)

    loadMods(event, 1, new ILoadCallback {
      override def onLoad(loadStage: LoadStage): Unit = progressBar.step(loadStage.getModMetadata.getName)
    })

    ProgressManager.pop(progressBar)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleServerAboutToStartClient(event: FMLServerAboutToStartEvent): Unit = {
    loadMods(event, 1, null)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleServerStartingClient(event: FMLServerStartingEvent): Unit = {
    loadMods(event, 1, null)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleServerStartedClient(event: FMLServerStartedEvent): Unit = {
    loadMods(event, 1, null)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleServerStoppingClient(event: FMLServerStoppingEvent): Unit = {
    loadMods(event, 1, null)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.CLIENT)
  def handleServerStoppedClient(event: FMLServerStoppedEvent): Unit = {
    loadMods(event, 1, null)
    dispatchLoadables(event, 1, null)
  }

  @SideOnly(Side.SERVER)
  def handleConstructionServer(event: FMLConstructionEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handlePreInitServer(event: FMLPreInitializationEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleInitServer(event: FMLInitializationEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handlePostInitServer(event: FMLPostInitializationEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleServerAboutToStartServer(event: FMLServerAboutToStartEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleServerStartingServer(event: FMLServerStartingEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleServerStartedServer(event: FMLServerStartedEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleServerStoppingServer(event: FMLServerStoppingEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  @SideOnly(Side.SERVER)
  def handleServerStoppedServer(event: FMLServerStoppedEvent): Unit = {
    loadMods(event, 2, null)
    dispatchLoadables(event, 2, null)
  }

  private def dispatchLoadables(event: FMLEvent, side: Int, callback: ILoadCallback): Unit = {
    LOADABLES.foreach(loadableEntry => {
      val mod = MODS.apply(loadableEntry._1)
      val loadStage = new LoadStage(event, EnumStageType.getTypeFromClass(event.getClass), mod)
      Loader.instance().setActiveModContainer(CONTAINERS.apply(loadableEntry._1))

      loadableEntry._2.foreach(loadable => side match {
        case 0 => loadable.load(loadStage)
        case 1 => loadable.loadClient(loadStage)
        case 2 => loadable.loadServer(loadStage)
        case _ => throw new ModloadingException(s"Wrong side [Side: $side] in dispatchLoadables!")
      })

      if (callback != null) callback.onLoad(loadStage)
    })
  }

  private def loadMods(event: FMLEvent, side: Int, callback: ILoadCallback): Unit = {
    MODS.foreach(modEntry => {
      val mod = modEntry._2
      val loadStage = new LoadStage(event, EnumStageType.getTypeFromClass(event.getClass), mod)
      Loader.instance().setActiveModContainer(CONTAINERS.apply(mod.getModid))

      side match {
        case 0 => {
          mod.load(loadStage)
          if (LOADABLES.contains(mod.getModid)) LOADABLES.apply(mod.getModid).appendAll(loadStage.getLoadablesAsBuffer)
          else {
            val buffer = new ArrayBuffer[ILoadable]()
            buffer.appendAll(loadStage.getLoadablesAsBuffer)
            LOADABLES.put(mod.getModid, buffer)
          }
        }
        case 1 => mod.getClientProxy.load(loadStage)
        case 2 => mod.getServerProxy.load(loadStage)
        case _ => throw new ModloadingException(s"Wrong side [Side: $side] in loadMods!")
      }

      if (callback != null) callback.onLoad(loadStage)
    })
  }

  private def loadMetadataJson(modid: String): (String, ModMetadata) = {
    val jsonObject = loadJson(new ResourceLocation(modid, "metadata"))
    val metadata = new ModMetadata
    metadata.setModid(modid)

    traverse(jsonObject.entrySet()).foreach(entry => {
      entry.getKey match {
        case "name" => metadata.setName(getString(jsonObject, "name"))
        case "version" => metadata.setVersion(getString(jsonObject, "version"))
        case "icon" => metadata.setIcon(getString(jsonObject, "icon"))
        case "authors" => traverse(getJsonObject(jsonObject, "authors").entrySet()).foreach(author => {
          val name = author.getKey
          val uuid = UUID.fromString(author.getValue.getAsJsonPrimitive.getAsString)
          metadata.addAuthor(new GameProfile(uuid, name))
        })
        case "website" => metadata.setWebsite(new URI(getString(jsonObject, "website")))
        case _ =>
      }
    })

    return (modid, metadata)
  }

  private def discoverMods(dataTable: ASMDataTable): ArrayBuffer[IMod] = {
    val mods = new ArrayBuffer[IMod]()
    val dataIterator = dataTable.getAll(classOf[Mod].getName).iterator()

    while (dataIterator.hasNext) {
      val modClass = Class.forName(dataIterator.next().getClassName)
      val modInstance = modClass.newInstance()

      modInstance match {
        case mod: IMod => mods += mod
        case _ => throw new ModloadingException(s"Class [${modClass.getName}] marked as mod candidate must be an instance of IMod!")
      }
    }

    mods
  }

}
